// (c) 2006, 2007 Nerdfever.com
// This version for "Rev3" hardware board (1st PCB version)

#ifndef __HARDWARE_H
#define __HARDWARE_H

#include <p18f2620.h>
#include "djllib.h"
#include "tweaks.h"

/*************************************************************
	INTERNAL PIC HARDWARE USE

	EEPROM		-
	FLASH		-	Used for data storage from FLASH_BASE
	Timer0		-
	Timer1		-	Generates RTC_INTERRUPT_INTERVAL_MS interrupts on overflow
	Timer2		-	Drives IR modulation thru CCP2
	Timer3		-	Times length of servo pulses via CCP1
	CCP1		-	Turns off servo pulse in Compare mode
	CPP2		-	IR output modulation, PWM mode
	MSSP		-
	EUSART		-	GPS I/O
	Comparator1	-	GPX data RX level converter
	Comparator2	-
	CVReference - 	Reference voltage for Comparator1
	HLDV 		-
**************************************************************/

// FLASH WRITING SUPPORT

#define FLASH_ROW_SIZE	(64)		// bytes, for PIC18F2620 (MUST be a power of 2, or fix WriteFlashStreaming())
#define FLASH_BUFFER_ROWS ((sizeof(LOG_STATUS_RECORD)*LAGGED_RECORDS)/FLASH_ROW_SIZE+2)		// rows, for "safe" writing to flash


// HARDWARE PORT SETUP

#define PORTA_IO	(0b00001011)
#define PORTB_IO	(0b11001111)
#define PORTC_IO	(0b10001000)

// HARDWARE ABSTRACTIONS

#define GPS_LED		(LED_YELLOW)		// flashes on GPS fix
#define WRITE_LED	(LED_RED)			// flash on write to flash memory

#define EEPROM_SIZE	(1024)				// bytes of EEPROM available

	// Outputs

#define LED_RED		(LATAbits.LATA5)	// red LED
#define LED_YELLOW	(LATBbits.LATB4)	// yellow LED
#define LED_GREEN	(LATBbits.LATB5)	// green LED

#define HORN_HW		(LATAbits.LATA6)	// loud horn (smoke alarm)
#define	PIEZO		(LATAbits.LATA7)	// piezo speaker (not so loud)

#define EJ_TRIGGER	(LATCbits.LATC0)	// ejection pyro charge trigger
#define SHUTTER		(LATCbits.LATC1)	// camera shutter button/IR LED output to camera
#define SPEEDO		(LATCbits.LATC4)	// "Z-80 speedometer" output, high=busy, low=idle

	// Inputs

#ifdef	ICD2
	#define	SW_RED	(0)					// switch does reset in debug
#else
	#define SW_RED	(!PORTEbits.RE3)	// switch (1=depressed)
#endif
#define SW_YELLOW	(!PORTBbits.RB2)	// switch (1=depressed)
#define SW_GREEN	(!PORTBbits.RB3)	// switch (1=depressed)

#define PRESSURE	// analog input pressure sensor
#define L_DETECT	(!PORTBbits.RB0)	// launch detector
#define EJ_CONT_OK	(PORTBbits.RB0)		// L_DETECT is reused as continuity check for EJ_TRIGGER
#define LIGHT		(!PORTBbits.RB1)	// photodetector 1=light
#define SW_ARMED	(PORTCbits.RC3)		// 1=ARMED, 0=SAFE


	// Camera stuff

#define	SHUTTER_IR_CODE	    (0b10101010101010001111111111111)	// bits will be sent right-to-left (Pentax)
//#define	SHUTTER_IR_CODE	(0b1000101010101010001111111111111)	// w/t (Pentax, works)

	// Servo limits

#ifdef SERVO_MG995
	#define	LEFT		(2.49)		// pulse width - milliseconds (on servo)
	#define RIGHT		(0.51)		// pulse width - milliseconds (on servo)
#endif


#ifdef SERVO_COX_GOLD_TRUCK
	#define	LEFT		(2.3)		// pulse width - milliseconds (on servo) (was 2.4)
	#define RIGHT		(1.3)		// pulse width - milliseconds (on servo) (avoid worn gear spot)
#endif

#ifdef SERVO_COX_SILVER
	#define	LEFT		(2.3)		// pulse width - milliseconds (on servo)
	#define RIGHT		(0.3)		// pulse width - milliseconds (on servo)
#endif

#ifdef SERVO_HOBBICO_CS35MG
	#define	LEFT		(2.4)		// pulse width - milliseconds (on servo)
	#define RIGHT		(0.8)		// pulse width - milliseconds (on servo)
#endif

// PRE-PROCESSOR CALCULATED VALUES

#define TMR_START	(65536 - MS_TO_UTICKS(RTC_INTERRUPT_INTERVAL_MS) + 6)	// start tick for interrupt period (+n ticks for ISR reset of Timer1)
#define IR_MODULATION_PERIOD_UTICKS	(FOSC/4/IR_MODULATION_FREQ_HZ-1)		// Osc ticks per infrared "on" modulation period

// MACROS

#define	MS_TO_UTICKS(ms)	((UINT16)((ms)*((FOSC/4.)/1000)/8))	// converts milliseconds to Timer1 ticks (in preprocessor)

#define	DI()	(INTCONbits.GIE = 0) 		// Disable Interrupts
#define EI()	(INTCONbits.GIE = 1) 		// Enable Interrupts

#define NOOP()	_asm NOP _endasm			// NOP instruction

#define BAUD_VALUE(bps)	((FOSC/4/bps)-1)	// value for baud rate generator

// (milliseconds) is floating-point and can be as little as 0.052 ms (at 2 MHz FOSC)
#define wait(milliseconds) __waitloop_internal((FOSC/2000000.0)*(milliseconds*17.81125))	// waits given time

#define IR_OFF	(CCP2CON = 0b00000000)		// turn off PWM modulation of IR output (revert to off)
#define IR_ON	(CCP2CON = 0b00001100)		// turn on PWM modulated IR output

// CPU CLOCK SETUP CONSTANTS

#define		OSCCON_8MHZ		0b01110000	// 5.45 mA draw @ 5v
#define		OSCCON_4MHZ		0b01100000	// 2.97 mA draw @ 5v
#define		OSCCON_2MHZ		0b01010000	// 1.83 mA draw @ 5v
#define		OSCCON_1MHZ		0b01000000	// 1.35 mA draw @ 5v
#define		OSCCON_500KHZ	0b00110000	// 0.95 mA draw @ 5v
#define		OSCCON_250KHZ	0b00100000	// 0.81 mA draw @ 5v
#define		OSCCON_125KHZ	0b00010000	// 0.57 mA draw @ 5v
#define		OSCCON_31KHZ	0b00000000	// 0.23 mA draw @ 5v

#if	FOSC == 8000000
	#define OSCVAL OSCCON_8MHZ
#endif

#if	FOSC == 4000000
	#define OSCVAL OSCCON_4MHZ
#endif

#if	FOSC == 2000000
	#define OSCVAL OSCCON_2MHZ
#endif

#if	FOSC == 1000000
	#define OSCVAL OSCCON_1MHZ
#endif

#if	FOSC == 500000
	#define OSCVAL OSCCON_500KHZ
#endif

#if	FOSC == 250000
	#define OSCVAL OSCCON_250KHZ
#endif

#if	FOSC == 125000
	#define OSCVAL OSCCON_125KHZ
#endif

#if	FOSC == 31000
	#define OSCVAL OSCCON_31KHZ
#endif

// FUNCTION PROTOTYPES

void WriteByte(UINT8 byte);
void SetSerial(UINT32 baudrate);
void WriteFlashBlock(UINT8 *RAMptr, UINT24 FLASHaddr);
void SetServoPos(UINT8 position);
UINT24 AccumulateADCsamples(UINT16 times);
void TakePicture(void);
void __waitloop_internal(UINT32 i);	// use wait() instead; this is for #define wait() use ONLY
void WriteToFlashBuffer(void *RAMptr, UINT24 bytes);
void FlushFlashBufferSafe(UINT8 maxWrites);
void ResetFlashPtrSafe(UINT24 flashAddr);
void SetupHardware(void);

#endif // __HARDWARE_H
