// (c) 2006 Nerdfever.com

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "djllib.h"
#include "hardware.h"
#include "nmeaGps.h"
#include "globals.h"

double intpart(double x)
{
	double result;

	modf(x, &result);

	return result;
}

double fracpart(double x)
{
	double intpart;

	return modf(x, &intpart);
}

// returns ASCII hex nibble of low order 4 bits
char nibble2hex(UINT8 value)
{
	value &= 0x0f;		// mask off all but low order 4 bits

	if (value < 10)
		return value + '0';
	else
		return value + ('A' - 10);
}

// returns value of single ASCII hex digit
UINT8 hex2nibble(char digit)
{
	if (digit < '9'+1)
		return digit-'0';

	return tolower(digit)-'a'+10;
}

// output command to NMEA GPS
void nmeaCommand(const rom char *cmd)
{
	int i = 0;
	int j;
	UINT8 cksum = 0;

	for (j=0; j<3; j++)				// send 3 times for redundancy
	{
		WriteByte(CR);
		WriteByte(LF);

		WriteByte(cmd[i++]);		// write '$' but don't include in checksum

		while (cmd[i])
		{
			WriteByte(cmd[i]);
			cksum ^= cmd[i++];		// simple XOR
		}

		WriteByte('*');
		WriteByte(nibble2hex(cksum >> 4));
		WriteByte(nibble2hex(cksum));

		WriteByte(CR);
		WriteByte(LF);
	}
}

UINT8 numUsedEncode(UINT8 satsUsed)
{
	if (satsUsed == 0)
		return 0;

	if (satsUsed < 4)
		return 1;

	if (satsUsed < 7)
		return 2;

	return 3;		// 7 or more
}

char fixModeASCII[] = "*ADESN";

UINT8 fixModeEncode(UINT8 fixMode)
{
	UINT8 i = 0;

	while (fixModeASCII[++i])
	{
		 if (fixModeASCII[i] == fixMode)
		 	return i;
	}

	return 0;
}

// on return, *diff has been filled out
void AbsoluteToDifferentialGPSFix(GPSFIX *abs, GPSFIX_D *diff)
{
	diff->rtc16					= abs->rtc16;
	diff->fixType				= abs->fixType;
	diff->valid 				= abs->valid;
	diff->numUsedCode			= numUsedEncode(abs->satsUsed);
	diff->fixModeCode			= fixModeEncode(abs->fixMode);
	diff->deltaLat10kPerMin		= abs->lat10kPerMin - LastValidLoggedFix.fix.lat10kPerMin;
	diff->deltaLon10kPerMin		= abs->lon10kPerMin - LastValidLoggedFix.fix.lon10kPerMin;
	diff->milliseconds			= abs->milliseconds;
	diff->groundSpeed5cmPerS	= abs->groundSpeedcmPerS / 5;
	diff->course256PerCircle	= (UINT8) (((UINT24) abs->course100PerDeg * 256 + 128) / 36000);
	diff->altCode				= (UINT16) ((abs->altitude10cmMSL + 5000) / 3);
}

// moves pointer to point at byte immediately after input string
void nextString(char** str_ptr)
{
	while(*(*str_ptr)++);
}

// parses NMEA $GPRMC sentance
void parseGPRMC(char* text_ptr, GPSFIX *fix_ptr)
{
	double degrees;

	text_ptr += sizeof("$GPRMC");							// skip to UTC time as hhmmss.fff

	fix_ptr->hour = (*text_ptr++-'0')*10+(*text_ptr++-'0');
	fix_ptr->min  = (*text_ptr++-'0')*10+(*text_ptr++-'0');
	fix_ptr->milliseconds = (UINT16) atoi(text_ptr)*1000;	// seconds
	text_ptr += 3;											// skip over "ss."
	fix_ptr->milliseconds += atoi(text_ptr);				// milliseconds

	nextString(&text_ptr);									// skip to status as A (valid) or V (not valid)
	fix_ptr->valid = (*text_ptr == 'A');

	nextString(&text_ptr);									// skip to latitude as ddmm.mmmm
	fix_ptr->lat10kPerMin = (INT32)(modf(atof(text_ptr) / 100.0, &degrees) * 1000000);	// minutes to 1/10000 minutes
	fix_ptr->lat10kPerMin += (INT32)(degrees * 600000);		// convert degrees to 1/10000 minutes

	nextString(&text_ptr);									// skip to N (north) or S (south)
	if (*text_ptr == 'S')
		fix_ptr->lat10kPerMin = -fix_ptr->lat10kPerMin;

	nextString(&text_ptr);									// skip to longitude as ddmm.mmmm
	fix_ptr->lon10kPerMin = (INT32) (modf(atof(text_ptr) / 100.0, &degrees) * 1000000);	// minutes to 1/10000 minutes
	fix_ptr->lon10kPerMin += (INT32) (degrees * 600000);	// convert degrees to 1/10000 minutes

	nextString(&text_ptr);									// skip to E (east) or W (west)
	if (*text_ptr == 'W')
		fix_ptr->lon10kPerMin = -fix_ptr->lon10kPerMin;

	nextString(&text_ptr);									// skip to ground speed in knots
	fix_ptr->groundSpeedcmPerS = (UINT24) (atof(text_ptr) / (KNOTSperMPS / 100));	// convert knots to 0.01 meters per second

	nextString(&text_ptr);									// skip to course in degrees
	fix_ptr->course100PerDeg = (UINT16)(atof(text_ptr)*100.);		// unit is 1/100 degree

	nextString(&text_ptr);									// skip to date as ddmmyy
	fix_ptr->day   = (*text_ptr++-'0')*10+(*text_ptr++-'0');
	fix_ptr->month = (*text_ptr++-'0')*10+(*text_ptr++-'0');
	fix_ptr->year  = (*text_ptr++-'0')*10+(*text_ptr++-'0');

	nextString(&text_ptr);									// skip E/W indicator
	nextString(&text_ptr);									// skip to fix mode

	fix_ptr->fixMode = *text_ptr++;							// ASCII fix mode code (see .h)
}

// parses NMEA $GPGGA sentance
void parseGPGGA(char* text_ptr, GPSFIX *fix_ptr)
{
	UINT8 satsUsed;

	text_ptr += sizeof("$GPGGA");	// skip to UTC time as hhmmss.sss
 	nextString(&text_ptr);			// skip to latitude
	nextString(&text_ptr);			// skip to N/S
	nextString(&text_ptr);			// skip to longitude
	nextString(&text_ptr);			// skip to E/W

	nextString(&text_ptr);			// skip to Position Fix Indicator: 0=no fix/invalid, 1=SPS valid, 2=SPS+Differential valid,
	fix_ptr->fixType = atoi(text_ptr);								// 3=PPS valid

	nextString(&text_ptr);			// skip to satellites used
	fix_ptr->satsUsed = min(15, atoi(text_ptr));

	nextString(&text_ptr);			// skip to HDOP

	nextString(&text_ptr);			// skip to altitude
	fix_ptr->altitude10cmMSL = (INT24) (atof(text_ptr) * 10.);
}


// checks if line starts with given prefix and has a valid NMEA checksum.  Returns TRUE if so, else FALSE.
UINT8 validNMEASentence(static rom const char *prefix, UINT8 prefix_length)
{
	UINT8 i = 1;
	UINT8 cksum = 0;
	UINT8 rxcksum;
	UINT8 byte;

	if (strncmppgm2ram((char*)NMEAbuffer[NMEAreadLine],(far rom const char *)prefix, prefix_length-1))
		return FALSE;		// prefix doesn't match

	do
	{
		byte = NMEAbuffer[NMEAreadLine][i++];

		if (byte == '*')
			break;

		if (byte)
			cksum ^= byte;		// simple XOR
		else
			cksum ^= ',';		// re-substitute comma for zeros
	}
	while (i < NMEA_BUFSIZE-2);

	if (byte != '*')
		return FALSE;			// got to end of buffer without seeing checksum - error

	rxcksum  = hex2nibble(NMEAbuffer[NMEAreadLine][i++])<<4;
	rxcksum += hex2nibble(NMEAbuffer[NMEAreadLine][i]);

	if (rxcksum != cksum)
		return FALSE;			// bad checksum

	return TRUE;
}

// NOTE: CheckForNewGPSFix() is written ASSUMING that the RMC message always comes AFTER the GGA message.
// Updates globals AnyFix ANY fix, even if marked "invalid"), and PositionFix. FALSE if no fix available.
// if on return AnyFix==TRUE, fix in LastGPSFix (global) is updated.
void CheckForNewGPSFix(void)
{
	LOG_ABSOLUTE_GPS_RECORD *temp_fix_ptr;

 	static rom const char RMC[] = "$GPRMC";
	static rom const char GGA[] = "$GPGGA";

//	static rom const char PMTK[] = "$PMTK";

	AnyFix = PositionFix = FALSE;

	if (NMEAlineCount)		// if data to process
	{

//		if (!strncmppgm2ram((char*)NMEAbuffer[NMEAreadLine],(far rom const char *)PMTK,sizeof(PMTK)-1))
//			retval = FALSE;

//		if (!strncmppgm2ram((char*)NMEAbuffer[NMEAreadLine],(far rom const char *)GGA,sizeof(GGA)-1))
		if (validNMEASentence(GGA,sizeof(GGA)-1))
			parseGPGGA((char*)NMEAbuffer[NMEAreadLine], &(NextGPSFixPtr->fix));

//		if (!strncmppgm2ram((char*)NMEAbuffer[NMEAreadLine],(far rom const char *)RMC,sizeof(RMC)-1))
		if (validNMEASentence(RMC,sizeof(RMC)-1))
		{
			NextGPSFixPtr->fix.rtc16 = (UINT16) GPSBurstStartRTC;
			NextGPSFixPtr->fix.spareBit = 0;

			parseGPRMC((char*)NMEAbuffer[NMEAreadLine], &(NextGPSFixPtr->fix));

			FixCount = max(FixCount, ++FixCount);

			temp_fix_ptr = LastGPSFixPtr;		// swap next and last fix pointers
			LastGPSFixPtr = NextGPSFixPtr;
			NextGPSFixPtr = temp_fix_ptr;

			AnyFix = TRUE;

			if ((LastGPSFixPtr->fix.valid) && (LastGPSFixPtr->fix.satsUsed >= 4))
			{
				PositionFix = TRUE;
				ValidFixCount = max(ValidFixCount, ++ValidFixCount);
			}
		}

		// finish processing line

		NMEAlineCount--;
		NMEAreadLine = (NMEAreadLine + 1) & (NMEA_LINES - 1);
	}
}

void InitGPS(void)
{

#ifdef GPS_EM406

	// This code for SiRF III chipset as implemented in EM-406 GPS

	SetSerial(4800);							// NMEA default of 4800,8,N,1

	wait(2000);									// wait for GPS to wake up

	nmeaCommand("$PSRF100,1,38400,8,1,0");		// SiRF GPS to 38400 bps

	SetSerial(38400);

	wait(1000);									// wait a bit for new baudrate to "take" in GPS

	nmeaCommand("$PSRF105,0");					// development data OFF
	nmeaCommand("$PSRF103,00,00,01,01");		// GGA on
	nmeaCommand("$PSRF103,01,00,00,01");		// GLL off
	nmeaCommand("$PSRF103,02,00,00,01");		// GSA off
	nmeaCommand("$PSRF103,03,00,00,01");		// GSV off
	nmeaCommand("$PSRF103,04,00,01,01");		// RMC on
	nmeaCommand("$PSRF103,05,00,00,01");		// VTG off
	nmeaCommand("$PSRF103,06,00,00,01");		// MSS off
	nmeaCommand("$PSRF103,08,00,00,01");		// ZDA off

#endif

#ifdef GPS_GTOP_G33

	// This code for MTK (5 Hz) chipset as implemented in GTOP G33 Bluetooth GPS

	// we assume GPS is ALREADY powered up when we get here (push the button)

	SetSerial(115200);

	// GPRMC and GPGGA only, once per fix time
	nmeaCommand("$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0");

	wait(1000);

	nmeaCommand("$PMTK313,1");				// search for SBAS satellites
	nmeaCommand("$PMTK301,2");				// turn on WAAS (preferred to SBAS)

	wait(1000);

	nmeaCommand("$PMTK220,400");			// set fix rate up to 200 ms (5 Hz)

	wait(1000);

	nmeaCommand("$PMTK300,400,0,0,0,0");	// set fix rate up to 200 ms (5 Hz)

#endif

#ifdef GPS_FGPMMOPA6

	// This code for MTK (5 Hz) chipset as implemented in GTOP FGPMMOPA6

	SetSerial(38400);
//	nmeaCommand("$PMTK390,1,5,115200");
//
//	nmeaCommand("$PMTK251,115200");			// set 115,200 bps

//	wait(1000);

	// GPRMC and GPGGA only, once per fix time
	nmeaCommand("$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0");
	nmeaCommand("$PMTK313,1");				// search for SBAS satellites
	nmeaCommand("$PMTK301,2");				// turn on WAAS (preferred to SBAS)


#endif

}
