#include "djllib.h"
#include "globals.h"
#include "peripherals.h"
#include "logging.h"
#include "stateMachine.h"
#include "tweaks.h"

UINT8  stateTransitionInProgress = FALSE;	// set to TRUE when entering a new state
UINT16 nextLogRTC16;						// holds future rtc value at which to log data again
UINT24 nextCameraRTC24;						// holds future rtc value at which to trigger camera shutter
UINT24 IDLEexpireRTC24;						// when to go from IDLE to ARM automatically (after TC smoothed pressure values settle down) 
UINT24 TIMEDLAUNCHexpireRTC24;				// when to go from FLIGHT PENDING to FLIGHT ASCENT (for ground testing)
UINT16 FLIGHTendEjectionRTC16;				// when to go from FLIGHT APOGEE to FLIGHT DESCENT

// this sets up "default" conditons for a new state - specific state entry proc will vary this to taste
// on return, FALSE means state isn't yet fully setup - calling routine needs to tweak things.
// on return, TRUE means stat is already fully setup - bail out since we're already in a state transition (bug prevention)
UINT8 settingUpNewState(void)
{
	if (stateTransitionInProgress)
		return TRUE;

	stateTransitionInProgress = TRUE;				// this is cleared only in state_machine()

	LED_RED 			= OFF;
	LED_YELLOW 			= OFF;
	LED_GREEN 			= OFF;

	Horn 				= OFF;
	Buzzer 				= OFF;
	Beep				= OFF;
	EJ_TRIGGER 			= OFF;

	return FALSE;
}

/*********** STATE TRANSITIONS ***************/

void enterIDLE(void)
{
	if (settingUpNewState())
		return;

	LED_RED 			= ON;
	State 				= IDLE;
	Substate			= NONE;

	IDLEexpireRTC24 = RTC24 + AUTO_ARM_PERIOD;
}

void enterARMED(void)
{
	if (settingUpNewState())
		return;

	LED_YELLOW 			= ON;
	State 				= ARMED;
	Substate			= SAFE;
}

void enterSubstateARMED_DANGER(void)
{
	if (settingUpNewState())
		return;

	LED_YELLOW 			= ON;
	Buzzer 				= ON;
	nextLogRTC16		= RTC16;
	nextCameraRTC24 	= RTC24;
	State 				= ARMED;
	Substate			= DANGER;

	InitializeLog(); 
}

void enterTIMEDLAUNCH(void)
{
	if (settingUpNewState())
		return;;

	LED_GREEN 			= ON;
	State 				= TIMEDLAUNCH;
	Substate			= NONE;
	
	nextLogRTC16		= RTC16;
	
	TIMEDLAUNCHexpireRTC24 = RTC24 + FLIGHT_SAFE_PERIOD;
}

void enterFLIGHT(void)
{
	if (settingUpNewState())
		return;

	LED_GREEN 			= ON;
	State 				= FLIGHT;
	Substate 			= ASCENT;

	FlushLaggedRecords();								// flush LAGGED_RECORDS log entries to Flash RAM

	PadPressure = SlowTCpressure;						// record pressure at pad (ground level pressure)

	TakePicture();										// take 1 frame with camera
	nextCameraRTC24 = RTC24 + FAST_CAM_PERIOD;
	
	DoLogStatus = TRUE;
	DoLogFix 	= AnyFix;
}

void enterSubstateFLIGHT_APOGEE(void)
{
	if (settingUpNewState())
		return;

	LED_GREEN 			= ON;
	EJ_TRIGGER 			= ON;							// trigger parachute ejection pyro charge
	FLIGHTendEjectionRTC16 = RTC16 + EJECT_DURATION;	// when to stop ejection current pulse
	State 				= FLIGHT;
	Substate 			= APOGEE;

	ApogeePressure		= FastTCpressure;				// pressure at apogee
}

void enterSubstateFLIGHT_DESCENT(void)
{
	if (settingUpNewState())
		return;

	LED_GREEN 			= ON;
	State 				= FLIGHT;
	Substate 			= DESCENT;
	nextLogRTC16		= RTC16;
}

void enterFULL(void)
{
	if (settingUpNewState())
		return;

	DoLogStatus			= FALSE;
	DoLogFix			= FALSE;
	
	LED_RED 			= ON;
	LED_GREEN 			= ON;
	State 				= FULL;
	// note - Substate retains it's previous value from FLIGHT (important if reaching FULL while in FLIGHT ASCENT)
}

void enterTEST(void)
{
	if (settingUpNewState())
		return;

	State 				= TEST;
	Substate			= NONE;
}

/*********** STATE FUNCTIONS ***************/

// These are executed once per LOOP_PERIOD when in the given state.

void idle(void)
{
	if (SW_ARMED)								// if arming switch is set
	{
		if (PositionFix)						// beep on good 3D fix
			Beep = ON;
		else
			Beep = OFF;
			
		if (YET24(IDLEexpireRTC24))				// auto-switch to ARMED state
			enterARMED();						// upon timeout
	}
	else
	{
		Beep = OFF;
		
		if (YET24(IDLEexpireRTC24))				// if idle timeout occurs with "safe" switch
			IDLEexpireRTC24 = RTC24 + MS(100);	// postpone response a little bit
	}
	
	if (SW_RED)									// check continuity while Red held down
	{
		LED_GREEN = EjectionContinuity();
		LED_YELLOW = !EjectionContinuity();
	}
	else
	{
		LED_GREEN = OFF;
		LED_YELLOW = OFF;
	}

	if (SW_YELLOW)
		enterARMED();

	if (SW_GREEN)
		enterTIMEDLAUNCH();
}

void armed(void)
{
	if (EjectionContinuity())						// yell if continuity no good
		Horn = OFF;
	else
		Horn = FAST_HORN;

	if (PositionFix)
		Beep = ON;
	else
		Beep = OFF;
	
	switch (Substate)
	{
		case SAFE:

			if (SW_ARMED)
				enterSubstateARMED_DANGER();		// arming switch is on - enter DANGER
			
			break;

		case DANGER:

			// operate camera

			if (YET24(nextCameraRTC24))				// if time for next camera frame
			{
				TakePicture();						// take 1 frame with camera
		
				nextCameraRTC24 += SLOW_CAM_PERIOD;
			}
				
			// check SAFE/DANGER switch for change

			if (!SW_ARMED)
				enterARMED();						// disarm into "safe" substate
			
			// check for launch

			if (LDet)								// detected launch
				enterFLIGHT();
			
			// do logging

			if (YET16(nextLogRTC16))				// if time to log
			{	
				nextLogRTC16 += SLOW_LOG_PERIOD;	// setup for next log entry

				if (FlashWritePtr > FLASH_BASE + FLASH_SIZE/4)
				{
					InitializeLog();				// if out of flash space (don't use more than 1/4 in ARMED)
				}
						
				DoLogStatus = TRUE;					// set logging flags
				DoLogFix 	= TRUE;
			}

			break;
	}

	if (SW_RED)
		enterIDLE();

	if (SwYellowHeld >= LONG_DOWN_PERIOD)
		enterTEST();

	if (SW_GREEN)
		enterTIMEDLAUNCH();
}

void timedlaunch(void)
{
	if (EjectionContinuity())			// yell if continuity no good
		Horn = OFF;
	else
		Horn = FAST_HORN;

	// we have to log in TIMEDLAUNCH or RTC16 rollover may occur and confuse the log decoder
	
	if (YET16(nextLogRTC16))				// if time to log
	{	
		nextLogRTC16 += SLOW_LOG_PERIOD;	// setup for next log entry
						
		DoLogStatus = TRUE;					// set logging flags
		DoLogFix 	= TRUE;
	}

	if (SW_ARMED)
	{
		// force immediate FLIGHT ASCENT if button held down
		if (YET24(TIMEDLAUNCHexpireRTC24) || (SwGreenHeld >= HOLD_DOWN_PERIOD))	
			enterFLIGHT();
	}
	else
		enterARMED();					// if not in Danger mode go back to ARMED

	if (SW_RED)
		enterIDLE();

	if (SW_YELLOW)
		enterARMED();
}

void flight(void)
{
	switch (Substate)
	{
		case ASCENT:

			DoLogFix = AnyFix;
			DoLogStatus = TRUE;
			Horn = SW_ARMED;

			if (ADet)
				enterSubstateFLIGHT_APOGEE();

			break;

		case APOGEE:

			DoLogFix = AnyFix;
			DoLogStatus = TRUE;
			Horn = SW_ARMED;

			if (YET16(FLIGHTendEjectionRTC16))
				enterSubstateFLIGHT_DESCENT();

			break;

		case DESCENT:
			
			DoLogFix = AnyFix;
			
			if (YET16(nextLogRTC16))					// if time to log
			{	
				nextLogRTC16 += MED_LOG_PERIOD;			// setup for next log entry
				DoLogStatus = TRUE;						// set logging flag
			}
			else
				DoLogStatus = FALSE;
							
			Horn = SLOW_HORN & ~(SW_ARMED-1);
						
			break;
	}

	// photos

	if (YET24(nextCameraRTC24))			// if time for next camera frame
	{
		TakePicture();					// take 1 frame with camera

		nextCameraRTC24 += FAST_CAM_PERIOD;
	}

	if (SW_RED)
		enterIDLE();

	if (SW_YELLOW)
		enterARMED();

	if (FlashWritePtr-FLASH_BASE >= FLASH_SIZE)		// if out of flash space
		enterFULL();
}

void full(void)
{
	BeepOutMaxAltitude();

	LastUARTrxRTC = RTC;			// trick to prevent flight() from writing to flash

	flight();						// do usual apogee detection & ejection

	// Note - DO NOT call enterFLIGHT() from here since the flash is full; you'll overrrun it.
}

void test(void)
{
	LaunchThreshold = TEST_THRESHOLD;	// for ground testing

	if (DeltaTC > LaunchThreshold)		// if going up
		LED_GREEN = ON;
	else
		LED_GREEN = OFF;

	if (DeltaTC < -LaunchThreshold)	// if going down
		LED_RED	= ON;
	else
		LED_RED	= OFF;

	if (SW_YELLOW)						// if yellow switch held down
		LED_YELLOW = LIGHT;				// led follows light detection
	else
		LED_YELLOW = !LED_YELLOW;		// otherwise, blink LED fast

	if (SW_RED)
		enterIDLE();

	if (SW_GREEN)
		enterTIMEDLAUNCH();
}

/*********** STATE MACHINE ***************/

void StateMachine(void)
{
	stateTransitionInProgress = FALSE;

	switch(State)
	{
		case IDLE:
			idle();
			break;

		case ARMED:
			armed();
			break;

		case TIMEDLAUNCH:
			timedlaunch();
			break;

		case FLIGHT:
			flight();
			break;

		case FULL:
			full();
			break;

		case TEST:
			test();
			break;
	}
}